/* -*- Mode: C;-*-
 *
 * This file is part of XDelta - A binary delta generator.
 *
 * Copyright (C) 1997, 1998  Josh MacDonald
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 *
 * xdmain.c,v 1.7 1998/10/03 02:32:09 jmacd Exp
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>  /* for ntohl, htonl on some */

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>

#ifndef _WIN32
#include <netinet/in.h> /* htonl on others */
#include <unistd.h>
#endif

#include <zlib.h>

#include "xdelta.h"
#include "xdmd5.h"

#define XD_PAGE_SIZE (1<<15)

#define XDELTA_100_PREFIX "%XDZ002%"
#define XDELTA_020_PREFIX "%XDZ001%"
#define XDELTA_018_PREFIX "%XDZ000%"
#define XDELTA_014_PREFIX "%XDELTA%"
#define XDELTA_PREFIX     XDELTA_100_PREFIX
#define XDELTA_PREFIX_LEN 8

#define HEADER_WORDS (6)
#define HEADER_SPACE (HEADER_WORDS*4)
/* The header is composed of 4-byte words (network byte order) as follows:
 * word 1: flags
 * word 2: (from name length) << 16 | (to name length)
 * word 3: (from file length)
 * word 4: (to file length)
 * word 5: (reserved)
 * word 6: (reserved)
 * flags are:
 */
#define FLAG_NO_VERIFY 1
#define FLAG_FROM_COMPRESSED 2
#define FLAG_TO_COMPRESSED 4
#define FLAG_PATCH_COMPRESSED 8
/* and, the header is follwed by the from file name, then the to file
 * name, then the data. */

#ifdef _WIN32
#define FOPEN_READ_ARG "rb"
#define FOPEN_WRITE_ARG "wb"
#define FILE_SEPARATOR '\\'
#else
#define FOPEN_READ_ARG "r"
#define FOPEN_WRITE_ARG "w"
#define FILE_SEPARATOR '/'
#endif

#include "getopt.h"

/* $Format: "static const char xdelta_version[] = \"$ReleaseVersion$\"; " $ */
static const char xdelta_version[] = "1.0.0"; 

typedef struct _Command Command;

struct _Command {
  gchar* name;
  gint (* func) (gint argc, gchar** argv);
  gint nargs;
  gint nvers;
};

static gint    delta_command    (gint argc, gchar** argv);
static gint    patch_command    (gint argc, gchar** argv);

static const Command commands[] =
{
  { "delta",    delta_command,    -1, 0 },
  { "patch",    patch_command,    -1, 0 },
  { NULL, NULL, 0, 0 }
};

static struct option const long_options[] =
{
  {"help",                no_argument, 0, 'h'},
  {"version",             no_argument, 0, 'v'},
  {"verbose",             no_argument, 0, 'V'},
  {"noverify",            no_argument, 0, 'n'},
  {"maxmem",              required_argument, 0, 'm'},
  {0,0,0,0}
};

static const gchar* program_name;
static gint         nvers = 0;
static gint         compress_level = Z_DEFAULT_COMPRESSION;
static gint         no_verify = FALSE;
static gint         verbose = FALSE;
static gint         max_mapped_pages = G_MAXINT;

#define xd_error g_warning

static void
usage ()
{
  xd_error ("usage: %s COMMAND [OPTIONS] [ARG1 ...]\n", program_name);
  xd_error ("use --help for more help\n");
  exit (2);
}

static void
help ()
{
  xd_error ("usage: %s COMMAND [OPTIONS] [ARG1 ARG2 ...]\n", program_name);
  xd_error ("COMMAND is one of:\n");
  xd_error ("  delta     Produce a delta from ARG1 to ARG2 producing ARG3\n");
  xd_error ("  patch     Patch file ARG1 with ARG2 producing ARG3\n");
  xd_error ("OPTIONS are:\n");
  /*  xd_error ("  -b, --base64\n");*/
  xd_error ("  -v, --version\n");
  xd_error ("  -V, --verbose\n");
  xd_error ("  -h, --help\n");
  xd_error ("  -n, --noverify\n");
  xd_error ("  -m, --maxmem=SIZE  Set the buffer size limit, e.g. 640K, 16M\n");
  xd_error ("  -[0-9]    Compression level: 0=none, 1=fast, 6=default, 9=best\n");
  exit (2);
}

static void
version ()
{
  xd_error ("version %s\n", xdelta_version);
  exit (2);
}

static gchar*
md5_as_hex (const guint8 *raw_md5)
{
  static gchar formatted_md5[33];
  gint i;

  for (i=0;i<16;i+=1)
    sprintf (formatted_md5 + i*2, "%02x", (0xff) & raw_md5[i]);

  formatted_md5[32] = 0;

  return formatted_md5;
}

static gchar*
strip_leading_path (gchar* p)
{
  gchar* ls = strrchr (p, FILE_SEPARATOR);

  if (ls)
    return ls + 1;

  return p;
}

static void
xd_error_func (const gchar   *log_domain,
	       GLogLevelFlags	log_level,
	       const gchar   *message,
	       gpointer	user_data)
{
  fprintf (stderr, "%s: %s", program_name, message);
}

gint
main(gint argc, gchar** argv)
{
  const Command *cmd = NULL;
  gint c;
  gint longind;

  program_name = strip_leading_path (argv[0]);

  g_log_set_handler (G_LOG_DOMAIN,
		     G_LOG_LEVEL_WARNING,
		     xd_error_func,
		     NULL);

  if (argc < 2)
    usage ();

  for (cmd = commands; cmd->name; cmd += 1)
    if (strcmp (cmd->name, argv[1]) == 0)
      break;

  if (strcmp (argv[1], "-h") == 0 ||
      strcmp (argv[1], "--help") == 0)
    help ();

  if (strcmp (argv[1], "-v") == 0 ||
      strcmp (argv[1], "--version") == 0)
    version ();

  if (!cmd->name)
    {
      xd_error ("unrecognized command\n");
      help ();
    }

  argc -= 1;
  argv += 1;

  while ((c = getopt_long(argc,
			  argv,
			  "+nhvVm:0123456789",
			  long_options,
			  &longind)) != EOF)
    {
      switch (c)
	{
	case 'n': no_verify = TRUE; break;
	case 'V': verbose = TRUE; break;
	case 'm':
	  {
	    gchar* end = NULL;
	    glong l = strtol (optarg, &end, 0);

	    if (end && g_strcasecmp (end, "M") == 0)
	      l <<= 20;
	    else if (end && g_strcasecmp (end, "K") == 0)
	      l <<= 10;
	    else if (end || l < 0)
	      {
		xd_error ("illegal maxmem argument %s\n", optarg);
		return 2;
	      }

	    l = MAX (l, XD_PAGE_SIZE * 8);

	    max_mapped_pages = l / XD_PAGE_SIZE;
	  }
	  break;
	case 'h': help (); break;
	case 'v': version (); break;
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	  compress_level = c - '0';
	  break;
	case '?':
	default:
	  xd_error ("illegal argument, use --help for help\n");
	  return 2;
	}
    }

  if (verbose && max_mapped_pages < G_MAXINT)
    xd_error ("using %d kilobytes of buffer space\n", (max_mapped_pages * XD_PAGE_SIZE) >> 10);

  argc -= optind;
  argv += optind;

  if (cmd->nargs >= 0 && argc != cmd->nargs)
    {
      xd_error ("wrong number of arguments\n");
      return 2;
    }

  if (nvers > cmd->nvers)
    {
      xd_error ("too many -r arguments\n");
      return 2;
    }

  return (* cmd->func) (argc, argv);
}

/* Commands */

#define READ_TYPE 1
#define READ_NOSEEK_TYPE 1
#define READ_SEEK_TYPE   3
#define WRITE_TYPE 4

typedef struct _LRU LRU;

struct _LRU
{
  LRU *next;
  LRU *prev;

  gint refs;
  guint page;
  guint8* buffer;
};

/* Note to the casual reader: the filehandle implemented here is
 * highly aware of the calling patterns of the Xdelta library.  It
 * also plays games with narrowing to implement several files in the
 * same handle.  So, if you try to modify or use this code, BEWARE.
 * See the repository package for a complete implementation.
 * In fact, its really quite a hack. */

struct _FileHandle
{
  guint    length;
  guint    real_length;
  gint     type;
  const char* name;
  const char* cleanup;

  guint8 md5[16];
  XdeltaMD5Ctx ctx;

  SerialSink   sink;
  SerialSource source;

  /* for write */
  int out_fd;
  void* out;
  gboolean (* out_write) (FileHandle* handle, const void* buf, gint nbyte);
  gboolean (* out_close) (FileHandle* handle);

  /* for read */
  GPtrArray *lru_table;
  LRU       *lru_head;  /* most recently used. */
  LRU       *lru_tail;  /* least recently used. */
  GMemChunk *lru_chunk;
  guint      lru_count;

  guint    narrow_low;
  guint    narrow_high;
  guint    current_pos;
  FILE*    in;
  gboolean (* in_read) (FileHandle* handle, void* buf, gint nbyte);
  gboolean (* in_close) (FileHandle* handle);
  gboolean in_compressed;

  const guint8* copy_page;
  guint  copy_pgno;
  gboolean md5_good;
  gboolean reset_length_next_write;

  gint md5_page;
  gint fd;
};

static gssize xd_handle_map_page (FileHandle *fh, guint pgno, const guint8** mem);
static gboolean xd_handle_unmap_page (FileHandle *fh, guint pgno, const guint8** mem);

static gboolean
xd_fwrite (FileHandle* fh, const void* buf, gint nbyte)
{
  return fwrite (buf, nbyte, 1, fh->out) == 1;
}

static gboolean
xd_fread (FileHandle* fh, void* buf, gint nbyte)
{
  return fread (buf, nbyte, 1, fh->in) == 1;
}

static gboolean
xd_fclose (FileHandle* fh)
{
  return fclose (fh->out) == 0;
}

static gboolean
xd_frclose (FileHandle* fh)
{
  return fclose (fh->in) == 0;
}

static gboolean
xd_gzwrite (FileHandle* fh, const void* buf, gint nbyte)
{
  return gzwrite (fh->out, (void*) buf, nbyte) == nbyte;
}

static gboolean
xd_gzread (FileHandle* fh, void* buf, gint nbyte)
{
  return gzread (fh->in, buf, nbyte) == nbyte;
}

static gboolean
xd_gzclose (FileHandle* fh)
{
  return gzclose (fh->out) == Z_OK;
}

static gboolean
xd_gzrclose (FileHandle* fh)
{
  return gzclose (fh->in) == Z_OK;
}

static void
init_table (FileHandle* fh)
{
  fh->lru_table = g_ptr_array_new ();
  fh->lru_chunk = g_mem_chunk_create(LRU, 1<<9, G_ALLOC_ONLY);
  fh->lru_head = NULL;
  fh->lru_tail = NULL;
}

static void
init_serialize (FileHandle* fh)
{
  ser_sink_init (&fh->sink);
  ser_source_init (&fh->source);
}

static FileHandle*
open_common (const char* name, const char* real_name)
{
  FileHandle* fh;

  gint fd;
  struct stat buf;

  if ((fd = open (name, O_RDONLY, 0)) < 0)
    {
      xd_error ("open %s failed: %s\n", name, g_strerror (errno));
      return NULL;
    }

  if (stat (name, &buf) < 0)
    {
      xd_error ("stat %s failed: %s\n", name, g_strerror (errno));
      return NULL;
    }

  fh = g_new0 (FileHandle, 1);

  fh->name = real_name;
  fh->fd = fd;
  fh->length = buf.st_size;
  fh->narrow_high = buf.st_size;

  return fh;
}

static gboolean
file_gzipped (const char* name, gboolean *is_compressed)
{
  FILE* f = fopen (name, FOPEN_READ_ARG);
  guint8 buf[2];

  (*is_compressed) = FALSE;

  if (! f)
    {
      xd_error ("open %s failed: %s\n", name, g_strerror (errno));
      return FALSE;
    }

  if (fread (buf, 2, 1, f) != 1)
    return TRUE;

#define GZIP_MAGIC1 037
#define GZIP_MAGIC2 0213

  if (buf[0] == GZIP_MAGIC1 && buf[1] == GZIP_MAGIC2)
    (* is_compressed) = TRUE;

  return TRUE;
}

static const char*
xd_tmpname (void)
{
  const char* tmpdir = getenv ("TMPDIR");
  GString* s;
  gint x = getpid ();
  static gint seq = 0;
  struct stat buf;

  if (! tmpdir)
    tmpdir = "/tmp";

  s = g_string_new (NULL);

  do
    {
      g_string_sprintf (s, "%s/xdtmp.%d.%d", tmpdir, x, seq++);
    }
  while (lstat (s->str, &buf) == 0);

  return s->str;
}

static const char*
file_gunzip (const char* name)
{
  const char* new_name = xd_tmpname ();
  FILE* out = fopen (new_name, FOPEN_WRITE_ARG);
  gzFile in = gzopen (name, "rb");
  guint8 buf[1024];
  int nread;

  while ((nread = gzread (in, buf, 1024)) > 0)
    {
      if (fwrite (buf, nread, 1, out) != 1)
	{
	  xd_error ("write %s failed (during uncompression): %s\n", new_name, g_strerror (errno));
	  return NULL;
	}
    }

  if (nread < 0)
    {
      xd_error ("gzread %s failed: %s\n", name, g_strerror (errno));
      return NULL;
    }

  gzclose (in);

  if (fclose (out))
    {
      xd_error ("close %s failed (during uncompression): %s\n", new_name, g_strerror (errno));
      return NULL;
    }

  return new_name;
}

static FileHandle*
open_read_noseek_handle (const char* name, gboolean* is_compressed, gboolean will_read)
{
  FileHandle* fh;
  const char* name0 = name;

  /* we _could_ stream-read this file if compressed, but it adds a
   * lot of complexity.  the library can handle it, just set the
   * length to (XDELTA_MAX_FILE_LEN-1) and make sure that the end
   * of file condition is set when on the last page.  However, I
   * don't feel like it. */
  if (! file_gzipped (name, is_compressed))
    return NULL;

  if ((* is_compressed) && ! (name = file_gunzip (name)))
    return NULL;

  if (! (fh = open_common (name, name0)))
    return NULL;

  fh->type = READ_NOSEEK_TYPE;

  init_serialize (fh);

  xdp_md5_init (&fh->ctx);

  if (*is_compressed)
    fh->cleanup = name;

  if (will_read)
    {
      if (! (fh->in = fdopen (dup (fh->fd), FOPEN_READ_ARG)))
	{
	  xd_error ("fdopen: %s\n", g_strerror (errno));
	  return NULL;
	}
      fh->in_read = &xd_fread;
      fh->in_close = &xd_frclose;
    }
  else
    {
      init_table (fh);
    }

  return fh;
}

static void
xd_read_close (FileHandle* fh)
{
  if (fh->cleanup)
    unlink (fh->cleanup);

  close (fh->fd);

  if (fh->in)
    (*fh->in_close) (fh);
}

static FileHandle*
open_read_seek_handle (const char* name, gboolean* is_compressed)
{
  FileHandle* fh;
  const char* name0 = name;

  if (! file_gzipped (name, is_compressed))
    return NULL;

  if ((* is_compressed) && ! (name = file_gunzip (name)))
    return NULL;

  if (! (fh = open_common (name, name0)))
    return NULL;

  fh->type = READ_SEEK_TYPE;

  if (*is_compressed)
    fh->cleanup = name;

  init_table (fh);
  init_serialize (fh);

  xdp_md5_init (&fh->ctx);

  return fh;
}

static FileHandle*
open_write_handle (int fd, const char* name)
{
  FileHandle* fh = g_new0 (FileHandle, 1);

  fh->out_fd = fd;
  fh->out_write = &xd_fwrite;
  fh->out_close = &xd_fclose;

  if (! (fh->out = fdopen (dup (fh->out_fd), FOPEN_WRITE_ARG)))
    {
      xd_error ("fdopen %s failed: %s\n", name, g_strerror (errno));
      return NULL;
    }

  fh->type = WRITE_TYPE;
  fh->name = name;

  init_serialize (fh);

  xdp_md5_init (&fh->ctx);

  return fh;
}

static gint
xd_begin_compression (FileHandle* fh)
{
  gint filepos;

  if (compress_level == 0)
    return fh->real_length;

  if (! (fh->out_close) (fh))
    {
      xd_error ("fdclose failed: %s\n", g_strerror (errno));
      return -1;
    }

  filepos = lseek (fh->out_fd, 0, SEEK_END);

  if (filepos < 0)
    {
      xd_error ("lseek failed: %s\n", g_strerror (errno));
      return -1;
    }

  fh->out = gzdopen (dup (fh->out_fd), "wb");
  fh->out_write = &xd_gzwrite;
  fh->out_close = &xd_gzclose;

  if (! fh->out)
    {
      xd_error ("gzdopen failed: %s\n", g_strerror (errno));
      return -1;
    }

  if (gzsetparams(fh->out, compress_level, Z_DEFAULT_STRATEGY) != Z_OK)
    {
      int foo;
      xd_error ("gzsetparams failed: %s\n", gzerror (fh->out, &foo));
      return -1;
    }

  return filepos;
}

static gboolean
xd_end_compression (FileHandle* fh)
{
  if (compress_level == 0)
    return TRUE;

  if (! (fh->out_close) (fh))
    {
      xd_error ("fdclose failed: %s\n", g_strerror (errno));
      return FALSE;
    }

  if (lseek (fh->out_fd, 0, SEEK_END) < 0)
    {
      xd_error ("lseek failed: %s\n", g_strerror (errno));
      return FALSE;
    }

  fh->out = fdopen (dup (fh->out_fd), FOPEN_WRITE_ARG);
  fh->out_write = &xd_fwrite;
  fh->out_close = &xd_fclose;

  if (! fh->out)
    {
      xd_error ("fdopen failed: %s\n", g_strerror (errno));
      return FALSE;
    }

  return TRUE;
}

static guint32 header_space[HEADER_WORDS];
static guint8  magic_buf[XDELTA_PREFIX_LEN];

static gssize
xd_handle_length (FileHandle *fh)
{
  if (fh->in_compressed)
    return fh->current_pos;
  else
    return fh->narrow_high - fh->narrow_low;
}

static gssize
xd_handle_pages (FileHandle *fh)
{
  g_assert (fh->type & READ_TYPE);
  return xd_handle_length (fh) / XD_PAGE_SIZE;
}

static gssize
xd_handle_pagesize (FileHandle *fh)
{
  g_assert (fh->type & READ_TYPE);
  return XD_PAGE_SIZE;
}

static gint
on_page (FileHandle* fh, guint pgno)
{
  if (pgno > xd_handle_pages (fh))
    return -1;

  if (pgno == xd_handle_pages (fh))
    return xd_handle_length (fh) % XD_PAGE_SIZE;

  return XD_PAGE_SIZE;
}

static gboolean
xd_handle_close (FileHandle *fh, gint ignore)
{
  /* this is really a reset for writable files */

  if (fh->type == WRITE_TYPE)
    {
      if (fh->reset_length_next_write)
	{
	  fh->reset_length_next_write = FALSE;
	  fh->length = 0;
	  fh->narrow_high = 0;
	}

      fh->reset_length_next_write = TRUE;
      xdp_md5_final (fh->md5, &fh->ctx);
      xdp_md5_init (&fh->ctx);
    }
  else if (fh->in)
    {
      xdp_md5_final (fh->md5, &fh->ctx);
      xdp_md5_init (&fh->ctx);
      fh->md5_good = FALSE;
    }

  return TRUE;
}

static const guint8*
xd_handle_checksum_md5 (FileHandle *fh)
{
  if (fh->in && ! fh->md5_good)
    {
      xdp_md5_final (fh->md5, &fh->ctx);
      fh->md5_good = TRUE;
    }
  else if (fh->type != WRITE_TYPE && !fh->in)
    {
      const guint8* page;

      while (fh->md5_page <= xd_handle_pages (fh))
	{
	  gint pgno = fh->md5_page;
	  gint onpage;

	  if ((onpage = xd_handle_map_page (fh, pgno, &page)) < 0)
	    return NULL;

	  if (pgno == fh->md5_page)
	    {
	      fh->md5_page += 1;
	      xdp_md5_update (&fh->ctx, page, onpage);

	      if (fh->md5_page > xd_handle_pages (fh))
		xdp_md5_final (fh->md5, &fh->ctx);
	    }

	  if (! xd_handle_unmap_page (fh, pgno, &page))
	    return NULL;
	}
    }

  return g_memdup (fh->md5, 16);
}

static gboolean
xd_handle_set_pos (FileHandle *fh, guint pos)
{
  if (fh->current_pos == pos + fh->narrow_low)
    return TRUE;

  if (pos + fh->narrow_low > fh->narrow_high)
    {
      xd_error ("unexpected EOF in %s\n", fh->name);
      return FALSE;
    }

  fh->current_pos = pos + fh->narrow_low;

  if (fseek (fh->in, fh->current_pos, SEEK_SET))
    {
      xd_error ("fseek failed: %s\n", g_strerror (errno));
      return FALSE;
    }

  return TRUE;
}

static gboolean
xd_handle_narrow (FileHandle* fh, guint low, guint high, gboolean compressed)
{
  if (high > fh->length)
    {
      xd_error ("unexpected EOF in %s\n", fh->name);
      return FALSE;
    }

  fh->narrow_low = low;
  fh->narrow_high = high;

  xdp_md5_init (&fh->ctx);

  if (compressed)
    {
      (* fh->in_close) (fh);

      if (lseek (fh->fd, low, SEEK_SET) < 0)
	{
	  xd_error ("lseek failed: %s\n", g_strerror (errno));
	  return FALSE;
	}

      fh->in = gzdopen (dup (fh->fd), "rb");
      fh->in_read =  &xd_gzread;
      fh->in_close = &xd_gzrclose;
      fh->in_compressed = TRUE;
      fh->current_pos = 0;

      if (! fh->in)
	{
	  xd_error ("gzdopen failed: %s\n", g_strerror (errno));
	  return -1;
	}
    }
  else
    {
      if (! xd_handle_set_pos (fh, 0))
	return FALSE;
    }

  return TRUE;
}

static guint
xd_handle_get_pos (FileHandle* fh)
{
  return fh->current_pos - fh->narrow_low;
}

static gssize
xd_handle_read (FileHandle *fh, guint8 *buf, gsize nbyte)
{
  if (nbyte == 0)
    return 0;

  /* This is trouble for compression. */
  /*if (nbyte + fh->current_pos > fh->narrow_high)
    {
      xd_error ("unexpected EOF in %s\n", fh->name);
      return -1;
    }*/

  if (! (fh->in_read) (fh, buf, nbyte))
    {
      xd_error ("read failed: %s\n", g_strerror (errno));
      return -1;
    }

  if (!no_verify)
    xdp_md5_update (&fh->ctx, buf, nbyte);

  fh->current_pos += nbyte;

  return nbyte;
}

static gboolean
xd_handle_write (FileHandle *fh, const guint8 *buf, gsize nbyte)
{
  g_assert (fh->type == WRITE_TYPE);

  if (fh->reset_length_next_write)
    {
      fh->reset_length_next_write = FALSE;
      fh->length = 0;
      fh->narrow_high = 0;
    }

  if (! no_verify)
    xdp_md5_update (&fh->ctx, buf, nbyte);

  if (! (*fh->out_write) (fh, buf, nbyte))
    {
      xd_error ("write failed: %s\n", g_strerror (errno));
      return FALSE;
    }

  fh->length += nbyte;
  fh->real_length += nbyte;
  fh->narrow_high += nbyte;

  return TRUE;
}

static gboolean
xd_handle_really_close (FileHandle *fh)
{
  g_assert (fh->type == WRITE_TYPE);

  if (! (* fh->out_close) (fh) || close (fh->out_fd) < 0)
    {
      xd_error ("write failed: %s\n", g_strerror (errno));
      return FALSE;
    }

  return TRUE;
}

static LRU*
pull_lru (FileHandle* fh, LRU* lru)
{
  if (lru->next && lru->prev)
    {
      lru->next->prev = lru->prev;
      lru->prev->next = lru->next;
    }
  else if (lru->next)
    {
      fh->lru_tail = lru->next;
      lru->next->prev = NULL;
    }
  else if (lru->prev)
    {
      fh->lru_head = lru->prev;
      lru->prev->next = NULL;
    }
  else
    {
      fh->lru_head = NULL;
      fh->lru_tail = NULL;
    }

  lru->next = NULL;
  lru->prev = NULL;

  return lru;
}

static gboolean
really_free_one_page (FileHandle* fh)
{
  LRU *lru;

  while (fh->lru_tail && fh->lru_tail->refs == 0)
    {
      gint to_unmap;

      lru = pull_lru (fh, fh->lru_tail);

      g_assert (lru->buffer);

      to_unmap = on_page (fh, lru->page);

      fh->lru_count -= 1;

      if (to_unmap > 0 && munmap (lru->buffer, to_unmap))
	{
	  xd_error ("munmap failed: %s\n", g_strerror (errno));
	  return FALSE;
	}

      lru->buffer = NULL;

      return TRUE;
    }

  return TRUE;
}

#if 0
static void
print_lru (FileHandle* fh)
{
  LRU* lru = fh->lru_head;

  for (; lru; lru = lru->prev)
    {
      g_print ("page %d buffer %p\n", lru->page, lru->buffer);

      if (! lru->prev && lru != fh->lru_tail)
	g_print ("incorrect lru_tail\n");
    }
}
#endif

static gboolean
make_lru_room (FileHandle* fh)
{
  if (fh->lru_count == max_mapped_pages)
    {
      g_assert (fh->lru_tail && fh->lru_tail->refs == 0);

      if (! really_free_one_page (fh))
	return FALSE;
    }

  g_assert (fh->lru_count < max_mapped_pages);

  return TRUE;
}

static gssize
xd_handle_map_page (FileHandle *fh, guint pgno, const guint8** mem)
{
  LRU* lru;
  guint to_map;

  g_assert (fh->type & READ_TYPE);

  if (fh->lru_table->len < (pgno + 1))
    {
      gint olen = fh->lru_table->len;

      g_ptr_array_set_size (fh->lru_table, pgno + 1);

      while (olen <= pgno)
	fh->lru_table->pdata[olen++] = NULL;
    }

  lru = fh->lru_table->pdata[pgno];

  if (! lru)
    {
      lru = g_chunk_new0 (LRU, fh->lru_chunk);
      fh->lru_table->pdata[pgno] = lru;
      lru->page = pgno;
    }
  else if (lru->buffer)
    {
      pull_lru (fh, lru);
    }

  lru->prev = fh->lru_head;
  lru->next = NULL;

  fh->lru_head = lru;

  if (lru->prev)
    lru->prev->next = lru;

  if (! fh->lru_tail)
    fh->lru_tail = lru;

  to_map = on_page (fh, pgno);

  if (on_page < 0)
    {
      xd_error ("unexpected EOF in %s\n", fh->name);
      return -1;
    }

  if (! lru->buffer)
    {
      if (! make_lru_room (fh))
	return -1;

      fh->lru_count += 1;

      if (to_map > 0)
	{
	  if (! (lru->buffer = mmap (NULL, to_map, PROT_READ, MAP_PRIVATE, fh->fd, pgno * XD_PAGE_SIZE)))
	    {
	      xd_error ("mmap failed: %s\n", g_strerror (errno));
	      return -1;
	    }
	}
      else
	{
	  lru->buffer = (void*) -1;
	}

      if (pgno == fh->md5_page)
	{
	  if (! no_verify)
	    xdp_md5_update (&fh->ctx, lru->buffer, to_map);
	  fh->md5_page += 1;

	  if (fh->md5_page > xd_handle_pages (fh))
	    xdp_md5_final (fh->md5, &fh->ctx);
	}
    }

  (*mem) = lru->buffer;

  lru->refs += 1;

  return to_map;
}

static gboolean
xd_handle_unmap_page (FileHandle *fh, guint pgno, const guint8** mem)
{
  LRU* lru;

  g_assert (fh->type & READ_TYPE);

  g_assert (pgno < fh->lru_table->len);

  lru = fh->lru_table->pdata[pgno];

  g_assert (lru && lru->refs > 0);

  g_assert (lru->buffer == (*mem));

  (*mem) = NULL;

  lru->refs -= 1;

  if (lru->refs == 0 && fh->type == READ_NOSEEK_TYPE)
    {
      pull_lru (fh, lru);

      lru->next = fh->lru_tail;
      if (lru->next) lru->next->prev = lru;
      lru->prev = NULL;
      fh->lru_tail = lru;

      if (! really_free_one_page (fh))
	return FALSE;
    }

  return TRUE;
}

static gboolean
xd_handle_copy (FileHandle *from, FileHandle *to, guint off, guint len)
{
  if (from->in)
    {
      guint8 buf[1024];

      /*if (! xd_handle_set_pos (from, off))
	return FALSE;*/

      while (len > 0)
	{
	  guint r = MIN (1024, len);

	  if (xd_handle_read (from, buf, r) != r)
	    return FALSE;

	  if (! xd_handle_write (to, buf, r))
	    return FALSE;

	  len -= r;
	}
    }
  else
    {
      while (len > 0)
	{
	  guint off_page = off / XD_PAGE_SIZE;
	  guint off_off = off % XD_PAGE_SIZE;

	  gint on = on_page (from, off_page);
	  guint rem;
	  guint copy;

	  if (on <= 0)
	    {
	      xd_error ("unexpected EOF in %s\n", from->name);
	      return FALSE;
	    }

	  rem = on - off_off;
	  copy = MIN (len, rem);

	  if (from->copy_pgno != off_page &&
	      from->copy_page &&
	      ! xd_handle_unmap_page (from, from->copy_pgno, &from->copy_page))
	    return FALSE;

	  from->copy_pgno = off_page;

	  if (xd_handle_map_page (from, off_page, &from->copy_page) < 0)
	    return FALSE;

	  if (! xd_handle_write (to, from->copy_page + off_off, copy))
	    return FALSE;

	  len -= copy;
	  off += copy;
	}
    }

  return TRUE;
}

static gboolean
xd_callback (XdeltaGenerator  *gen,
	     void             *user_data,
	     const guint8     *expected_md5,
	     gboolean          required,
	     guint             offset,
	     guint             length,
	     XdeltaOutStream  *write_to)
{
  return xd_handle_copy (user_data, write_to, offset, length);
}

static gboolean
xd_handle_putui (FileHandle *fh, guint32 i)
{
  guint32 hi = htonl (i);

  return xd_handle_write (fh, (guint8*)&hi, 4);
}

static gboolean
xd_handle_getui (FileHandle *fh, guint32* i)
{
  if (xd_handle_read (fh, (guint8*)i, 4) != 4)
    return FALSE;

  *i = ntohl (*i);

  return TRUE;
}

static HandleFuncTable xd_handle_table =
{
  xd_handle_length,
  xd_handle_pages,
  xd_handle_pagesize,
  xd_handle_map_page,
  xd_handle_unmap_page,
  xd_handle_checksum_md5,

  xd_handle_close,

  xd_handle_write,
  xd_handle_copy,

  xd_handle_getui,
  xd_handle_putui,
  xd_handle_read
};

static XdeltaMD5Table xd_md5_table =
{
  /* These are only needed for rsync */
  /*md5_init*/ NULL,
  /*md5_update*/ NULL,
  /*md5_final*/ NULL,
  /*md5_ctx_size*/ 0
};

static void
htonl_array (guint32* array, gint len)
{
  gint i;

  for (i = 0; i < len; i += 1)
    array[i] = htonl(array[i]);
}

static void
ntohl_array (guint32* array, gint len)
{
  gint i;

  for (i = 0; i < len; i += 1)
    array[i] = ntohl(array[i]);
}

static gint
delta_command (gint argc, gchar** argv)
{
  gint patch_out_fd;
  const char* patch_out_name;
  FileHandle *from, *to, *out;
  XdeltaGenerator* gen;
  XdeltaSource* src;
  XdeltaControl* cont;
  gboolean from_is_compressed = FALSE, to_is_compressed = FALSE;
  guint32 control_offset, header_offset;
  const char* from_name, *to_name;

  /* This could read tofile from stdin... */
  if (argc < 2 || argc > 3)
    {
      xd_error ("usage: %s delta fromfile tofile [patchfile]\n", program_name);
      return 2;
    }

  if (! (from = open_read_seek_handle (argv[0], &from_is_compressed)))
    return 2;

  if (! (to = open_read_noseek_handle (argv[1], &to_is_compressed, FALSE)))
    return 2;

  if (argc == 2 || strcmp (argv[2], "-") == 0)
    {
      patch_out_fd = STDOUT_FILENO;
      patch_out_name = "standard output";
    }
  else
    {
      int fd = open (argv[2], O_WRONLY | O_CREAT | O_TRUNC, 0666);

      if (fd < 0)
	{
	  xd_error ("open %s failed: %s\n", argv[2], g_strerror (errno));
	  return 2;
	}

      patch_out_fd = fd;
      patch_out_name = argv[2];
    }

  if (! (out = open_write_handle (patch_out_fd, patch_out_name)))
    return 2;

  gen = xdp_generator_new (&xd_handle_table, &xd_md5_table);

  if (! (src = xdp_source_new (gen, NULL, from, NULL, NULL)))
    {
      xd_error ("source failed: %s\n", xdp_strerror (xdp_errno (gen)));
      return 2;
    }

  xdp_source_add (gen, src);

  if (! xd_handle_write (out, XDELTA_PREFIX, XDELTA_PREFIX_LEN))
    return 2;

  /* compute the header */
  header_space[0] = 0;

  if (no_verify) header_space[0]           |= FLAG_NO_VERIFY;
  if (from_is_compressed) header_space[0]  |= FLAG_FROM_COMPRESSED;
  if (to_is_compressed) header_space[0]    |= FLAG_TO_COMPRESSED;
  if (compress_level != 0) header_space[0] |= FLAG_PATCH_COMPRESSED;

  from_name = strip_leading_path (argv[0]);
  to_name = strip_leading_path (argv[1]);

  header_space[1] = strlen (from_name) << 16 | strlen (to_name);
  header_space[2] = xd_handle_length (from);
  header_space[3] = xd_handle_length (to);
  /* end compute the header */

  htonl_array (header_space, HEADER_WORDS);

  if (! xd_handle_write (out, (guint8*) header_space, HEADER_SPACE))
    return 2;

  if (! xd_handle_write (out, from_name, strlen (from_name)))
    return 2;

  if (! xd_handle_write (out, to_name, strlen (to_name)))
    return 2;

  if (! xd_handle_close (out, 0))
    return 2;

  if ((header_offset = xd_begin_compression (out)) < 0)
    return 2;

  if (! (cont = xdp_generate_delta (gen, to, FALSE, NULL, out)))
    {
      xd_error ("generate delta failed: %s\n", xdp_strerror (xdp_errno (gen)));
      return 2;
    }

  if (! xd_handle_close (out, 0))
    return 2;

  if ((control_offset = xd_begin_compression (out)) < 0)
    return 2;

  if (! xdp_control_write (gen, cont, out))
    {
      xd_error ("write delta control failed: %s\n", xdp_strerror (xdp_errno (gen)));
      return 2;
    }

  if (! xd_end_compression (out))
    return 2;

  if (! xd_handle_putui (out, control_offset))
    return 2;

  xd_read_close (from);
  xd_read_close (to);

  if (! xd_handle_really_close (out))
    return 2;

  return control_offset != header_offset;
}

static gint
patch_command (gint argc, gchar** argv)
{
  FileHandle* patch_in, *from_in, *to_out;
  int to_out_fd;
  gint16 from_name_len, to_name_len;
  gint32 expected_from_len, expected_to_len;
  gchar* from_name, *to_name;
  gboolean from_is_compressed = FALSE, patch_is_compressed = FALSE;
  guint patch_flags;
  guint control_offset, header_offset;
  XdeltaControl* cont;
  XdeltaGenerator* gen = xdp_generator_new (&xd_handle_table, &xd_md5_table);

  if (argc < 1 || argc > 3)
    {
      xd_error ("usage: %s patch patchfile fromfile [tofile]\n", program_name);
      return 2;
    }

  /* Strictly speaking, I'm violating the intended semantics of noseek here.
   * It will seek the file, which is not in fact checked in the map/unmap
   * logic above.  This only means that it will not cache pages of this file
   * since it will be read piecewise sequentially. */
  if (! (patch_in = open_read_noseek_handle (argv[0], &patch_is_compressed, TRUE)))
    return 2;

  if (xd_handle_read (patch_in, magic_buf, XDELTA_PREFIX_LEN) != XDELTA_PREFIX_LEN)
    return 2;

  if (xd_handle_read (patch_in, (guint8*) header_space, HEADER_SPACE) != HEADER_SPACE)
    return 2;

  ntohl_array (header_space, HEADER_WORDS);

  if (strncmp (magic_buf, XDELTA_PREFIX, XDELTA_PREFIX_LEN) != 0)
    {
      xd_error ("bad magic number: not a valid delta\n");
      return 2;
    }

  patch_flags = header_space[0];

  if (no_verify)
    xd_error ("--noverify is only accepted when creating a delta\n");

  if (patch_flags & FLAG_NO_VERIFY)
    no_verify = TRUE;
  else
    no_verify = FALSE;

  from_name_len = header_space[1] >> 16;
  to_name_len = header_space[1] & 0xffff;

  expected_from_len = header_space[2];
  expected_to_len = header_space[3];

  from_name = g_malloc (from_name_len+1);
  to_name = g_malloc (to_name_len+1);

  from_name[from_name_len] = 0;
  to_name[to_name_len] = 0;

  if (xd_handle_read (patch_in, from_name, from_name_len) != from_name_len)
    return 2;

  if (xd_handle_read (patch_in, to_name, to_name_len) != to_name_len)
    return 2;

  if (argc > 1)
    from_name = argv[1];
  else if (verbose)
    xd_error ("using default from file name: %s\n", from_name);

  if (argc > 2)
    to_name = argv[2];
  else if (verbose)
    xd_error ("using default to file name: %s\n", to_name);

  if (strcmp (to_name, "-") == 0)
    {
      to_out_fd = STDOUT_FILENO;
      to_name = "standard output";
    }
  else
    {
      to_out_fd = open (to_name, O_WRONLY | O_CREAT | O_TRUNC, 0666);

      if (to_out_fd < 0)
	{
	  xd_error ("open %s failed: %s\n", to_name, g_strerror (errno));
	  return 2;
	}
    }

  to_out = open_write_handle (to_out_fd, to_name);

  if ((patch_flags & FLAG_TO_COMPRESSED) && (xd_begin_compression (to_out) < 0))
    return 2;

  if (! (from_in = open_read_seek_handle (from_name, &from_is_compressed)))
    return 2;

  if (from_is_compressed != ((patch_flags & FLAG_FROM_COMPRESSED) && 1))
    xd_error ("warning: expected %scompressed from file\n", (patch_flags & FLAG_FROM_COMPRESSED) ? "" : "un");

  if (xd_handle_length (from_in) != expected_from_len)
    {
      xd_error ("expected from file of %slength %d bytes\n", from_is_compressed ? "uncompressed " : "", expected_from_len);
      return 2;
    }

  header_offset = xd_handle_get_pos (patch_in);

  if (! xd_handle_set_pos (patch_in, xd_handle_length (patch_in) - 4))
    return 2;

  if (! xd_handle_getui (patch_in, &control_offset))
    return 2;

  if (! xd_handle_narrow (patch_in, control_offset, xd_handle_length (patch_in) - 4, patch_flags & FLAG_PATCH_COMPRESSED))
    return 2;

  if (! (cont = xdp_control_read (gen, patch_in)))
    {
      xd_error ("invalid delta control: %s\n", xdp_strerror (xdp_errno (gen)));
      return 2;
    }

  if (! xd_handle_narrow (patch_in, header_offset, control_offset, patch_flags & FLAG_PATCH_COMPRESSED))
    return 2;

  if (! xdp_apply_delta (gen, cont, patch_in, xd_callback, from_in, to_out))
    {
      /* the from MD5 has not been checked yet, since it may not have been
       * read completely, and if the patch is okay there is no need to.
       * Here, test that because it is sure to cause failure. */

      if (! no_verify)
	{
	  const guint8 *fmd5 = xd_handle_checksum_md5 (from_in);
	  const guint8 *efmd5 = xdp_control_depends_on (cont, 0);

	  if (memcmp (fmd5, efmd5, 16) != 0)
	    {
	      xd_error ("mismatched from MD5 checksum, expected: %s\n", md5_as_hex (efmd5));
	      return 2;
	    }
	}

      xd_error ("apply delta failed: %s\n", xdp_strerror (xdp_errno (gen)));
      return 2;
    }

  xd_read_close (from_in);
  xd_read_close (patch_in);

  if (! xd_handle_really_close (to_out))
    return 2;

  return 0;
}
